package slimeknights.tconstruct.library.modifiers;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;

import net.minecraft.enchantment.Enchantment;
import net.minecraft.enchantment.EnchantmentHelper;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.SharedMonsterAttributes;
import net.minecraft.entity.ai.attributes.AttributeModifier;
import net.minecraft.entity.ai.attributes.IAttributeInstance;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.nbt.NBTTagString;
import net.minecraft.util.DamageSource;
import net.minecraft.util.StatCollector;

import java.util.Arrays;
import java.util.List;
import java.util.Random;

import slimeknights.mantle.util.RecipeMatchRegistry;
import slimeknights.tconstruct.library.TinkerRegistry;
import slimeknights.tconstruct.library.Util;
import slimeknights.tconstruct.library.utils.TagUtil;
import slimeknights.tconstruct.library.utils.TinkerUtil;

public abstract class Modifier extends RecipeMatchRegistry implements IModifier {

  public static final String LOC_Name = "modifier.%s.name";
  public static final String LOC_Desc = "modifier.%s.desc";
  public static final String LOC_Extra = "modifier.%s.extra";

  protected static final Random random = new Random();

  public final String identifier;

  protected final List<ModifierAspect> aspects = Lists.newLinkedList();

  public Modifier(String identifier) {
    this.identifier = Util.sanitizeLocalizationString(identifier);

    TinkerRegistry.registerModifier(this);
  }

  @Override
  public String getIdentifier() {
    return identifier;
  }

  @Override
  public boolean isHidden() {
    return false;
  }

  protected void addAspects(ModifierAspect... aspects) {
    this.aspects.addAll(Arrays.asList(aspects));
  }

  @Override
  public final boolean canApply(ItemStack stack, ItemStack original) throws TinkerGuiException {
    // aspects
    for(ModifierAspect aspect : aspects) {
      if(!aspect.canApply(stack, original)) {
        return false;
      }
    }

    NBTTagList modifiers = TagUtil.getBaseModifiersTagList(stack);
    for(int i = 0; i < modifiers.func_74745_c(); i++) {
      String id = modifiers.func_150307_f(i);
      IModifier mod = TinkerRegistry.getModifier(id);
      if(mod != null && !canApplyTogether(mod)) {
        throw new TinkerGuiException(Util.translateFormatted("gui.error.incompatible_modifiers", this.getLocalizedName(), mod.getLocalizedName()));
      }
    }

    for(int id : EnchantmentHelper.func_82781_a(stack).keySet()) {
      if(!canApplyTogether(Enchantment.func_180306_c(id))) {
        String enchName = StatCollector.func_74838_a(Enchantment.func_180306_c(id).func_77320_a());
        throw new TinkerGuiException(Util.translateFormatted("gui.error.incompatible_enchantments", this.getLocalizedName(), enchName));
      }
    }

    return canApplyCustom(stack);
  }

  public boolean canApplyTogether(Enchantment enchantment) {
    return true;
  }

  public boolean canApplyTogether(IModifier otherModifier) {
    return true;
  }

  protected boolean canApplyCustom(ItemStack stack) throws TinkerGuiException {
    return true;
  }

  @Override
  public void updateNBT(NBTTagCompound modifierTag) {
    // nothing to do in most cases, aspects handle the updating for most modifier
  }

  @Override
  public void apply(ItemStack stack) {
    NBTTagCompound root = TagUtil.getTagSafe(stack);
    apply(root);
    stack.func_77982_d(root);
  }

  @Override
  public void apply(NBTTagCompound root) {
    // add the modifier to its data
    NBTTagList tagList;

    // if the modifier wasn't present before, add it and safe it to the tool
    if(!TinkerUtil.hasModifier(root, getIdentifier())) {
      tagList = TagUtil.getBaseModifiersTagList(root);
      tagList.func_74742_a(new NBTTagString(getIdentifier()));
      TagUtil.setBaseModifiersTagList(root, tagList);
    }

    // have the modifier itself save its data
    NBTTagCompound modifierTag = new NBTTagCompound();
    tagList = TagUtil.getModifiersTagList(root);
    int index = TinkerUtil.getIndexInList(tagList, identifier);
    if(index >= 0) {
      modifierTag = tagList.func_150305_b(index);
    }

    // update NBT through aspects
    for(ModifierAspect aspect : aspects) {
      aspect.updateNBT(root, modifierTag);
    }

    updateNBT(modifierTag);

    // some modifiers might not save data, don't save them
    if(!modifierTag.func_82582_d()) {
      // but if they do, ensure that the identifier is correct
      ModifierNBT data = ModifierNBT.readTag(modifierTag);
      if(!identifier.equals(data.identifier)) {
        data.identifier = identifier;
        data.write(modifierTag);
      }
    }

    // update the tools NBT
    if(index >= 0) {
      tagList.func_150304_a(index, modifierTag);
    }
    else {
      tagList.func_74742_a(modifierTag);
    }

    TagUtil.setModifiersTagList(root, tagList);

    applyEffect(root, modifierTag);
  }

  @Override
  public String getTooltip(NBTTagCompound modifierTag, boolean detailed) {
    StringBuilder sb = new StringBuilder();

    ModifierNBT data = ModifierNBT.readTag(modifierTag);

    sb.append(getLocalizedName());
    if(data.level > 1) {
      sb.append(" ");
      sb.append(TinkerUtil.getRomanNumeral(data.level));
    }

    return sb.toString();
  }

  public String getLeveledTooltip(NBTTagCompound modifierTag, boolean detailed) {
    // the most important function in the whole file!
    ModifierNBT data = ModifierNBT.readInteger(modifierTag);

    String basic = getLocalizedName(); // backup
    if(data.level == 0) {
      return basic;
    }
    else if(data.level > 1) {
      basic += " " + TinkerUtil.getRomanNumeral(data.level);
    }

    for(int i = data.level; i > 1; i--) {
      if(StatCollector.func_94522_b(String.format(LOC_Name + i, getIdentifier()))) {
        basic = StatCollector.func_74838_a(String.format(LOC_Name + i, getIdentifier()));
        break;
      }
    }

    if(detailed) {
      return basic + " " + data.extraInfo;
    }

    return basic;
  }

  @Override
  public String getLocalizedName() {
    return Util.translate(LOC_Name, getIdentifier());
  }

  @Override
  public String getLocalizedDesc() {
    return Util.translate(LOC_Desc, getIdentifier());
  }

  @Override
  public List<String> getExtraInfo(ItemStack tool, NBTTagCompound modifierTag) {
    return ImmutableList.of();
  }

  @Override
  public boolean hasTexturePerMaterial() {
    return false;
  }

  protected static boolean attackEntitySecondary(DamageSource source, float damage, Entity entity, boolean ignoreInvulv, boolean resetInvulv) {
    return attackEntitySecondary(source, damage, entity, ignoreInvulv, resetInvulv, true);
  }

  protected static boolean attackEntitySecondary(DamageSource source, float damage, Entity entity, boolean ignoreInvulv, boolean resetInvulv, boolean noKnockback) {
    IAttributeInstance knockbackAttribute = null;
    float oldLastDamage = 0;
    if(entity instanceof EntityLivingBase) {
      oldLastDamage = ((EntityLivingBase) entity).field_110153_bc;
      if(noKnockback) {
        knockbackAttribute = ((EntityLivingBase) entity).func_110148_a(SharedMonsterAttributes.field_111266_c);
      }
    }

    if(knockbackAttribute != null) {
      knockbackAttribute.func_111121_a(ANTI_KNOCKBACK_MOD);
    }

    // set hurt resistance time to 0 because we always want to deal damage in traits
    if(ignoreInvulv) {
      entity.field_70172_ad = 0;
    }
    boolean hit = entity.func_70097_a(source, damage);
    if(entity instanceof EntityLivingBase) {
      ((EntityLivingBase) entity).field_110153_bc += oldLastDamage;
    }
    // reset hurt resistance time if desired
    if(hit && resetInvulv) {
      entity.field_70172_ad = 0;
    }

    if(knockbackAttribute != null) {
      knockbackAttribute.func_111124_b(ANTI_KNOCKBACK_MOD);
    }

    return hit;
  }

  private static final AttributeModifier ANTI_KNOCKBACK_MOD = new AttributeModifier("Anti Modifier Knockback", 1f, 0);
}
